/**
 * GBA application wide constants and macro definitions
 *
 * @author Ben Heiskell
 * @date 2009 08
 */

/* Ensure inclusion only occurs once */
#ifndef GBA_H
#define GBA_H

/* Following Linux naming conventions for unsigned ints */
typedef unsigned long  int uint32_t;
typedef unsigned short int uint16_t;

#define VIDBUFADR (uint16_t*)0x6000000
#define REG_DISPCTL (*(uint32_t*)0x4000000)

#define SCANLINE_COUNTER (*(volatile uint16_t *)0x4000006)
#define SCANLINE_OUT_OF_RANGE 160

#define BG2_ENABLE (1<<10)
#define MODE3 3 

#define PIXEL3(x,y) ((y) * S_WIDTH + (x))
#define RGB(r,g,b) ((r) | (g) << 5 | (b) << 10)

#define BLACK   RGB(0,0,0) 
#define WHITE   RGB(31,31,31)
#define RED     RGB(31,0,0)
#define GREEN   RGB(0,31,0)
#define BLUE    RGB(0,0,31)
#define CYAN    RGB(0,31,31)
#define MAGENTA RGB(31,0,31)
#define YELLOW  RGB(31,31,0)

#define S_WIDTH 240
#define S_HEIGHT 160

#define BUTTONS (*(volatile unsigned int *)0x04000130)

#define BUTTON_A       (1<<0)
#define BUTTON_B       (1<<1)
#define BUTTON_SELECT  (1<<2)
#define BUTTON_START   (1<<3)
#define BUTTON_RIGHT   (1<<4)
#define BUTTON_LEFT    (1<<5)
#define BUTTON_UP      (1<<6)
#define BUTTON_DOWN    (1<<7)
#define BUTTON_R       (1<<8)
#define BUTTON_L       (1<<9)

/* KEY_ACTIVE is a helper macro for DOWN and HIT */
#define KEY_ACTIVE(key, buttons) (~(buttons) & (key))
#define KEY_DOWN(key) (KEY_ACTIVE(key, BUTTONS))
#define KEY_HIT(key, last_butts) (KEY_DOWN(key) & ~KEY_ACTIVE(key, last_butts))

/**
 * Macro used to pause the game play. It makes the pause clear by fading out the
 * background. Don't try to decifer the pointer arithmetic used to do this. It
 * is a lot of bitshifting and masking to get the effect of dividing each color
 * by four.
 *
 * This was made a macro because as of now it doesn't fit anywhere else.
 */
#define PAUSE_GAME3() {\
  while(KEY_DOWN(BUTTON_START));\
  uint16_t *pg_i;\
  for (pg_i = VIDBUFADR; pg_i <= (VIDBUFADR + S_WIDTH * S_HEIGHT); pg_i++) {\
    *pg_i = (*pg_i&0x1F)>>2|((*pg_i)>>5&0x1F)>>2<<5|((*pg_i)>>10&0x1F)>>2<<10;\
  }\
  while(!KEY_DOWN(BUTTON_START));\
}

/* DMA */
typedef struct {
  const volatile void *src;
  volatile void *dst;
  volatile uint32_t cnt;
} DMAREC;

#define DMA ((volatile DMAREC *)0x040000B0)

#define DMA_DESTINATION_INCREMENT (0 << 21)
#define DMA_DESTINATION_DECREMENT (1 << 21)
#define DMA_DESTINATION_FIXED (2 << 21)
#define DMA_DESTINATION_RESET (3 << 21)

#define DMA_SOURCE_INCREMENT (0 << 23)
#define DMA_SOURCE_DECREMENT (1 << 23)
#define DMA_SOURCE_FIXED (2 << 23)

#define DMA_REPEAT (1 << 25)

#define DMA_16 (0 << 26)
#define DMA_32 (1 << 26)

#define DMA_NOW (0 << 28)
#define DMA_AT_VBLANK (1 << 28)
#define DMA_AT_HBLANK (2 << 28)
#define DMA_AT_REFRESH (3 << 28)

#define DMA_IRQ (1 << 30)
#define DMA_ON (1 << 31)

#define START_ON_FIFO_EMPTY 0x30000000

#endif
