/**
 * 
 *
 * @author Ben Heiskell
 * @date 2009 
 */

#include <stdlib.h>
#include <stdio.h>
#include "gba.h"
#include "font.h"
#include "shape.h"
#include "play_bg.h"
#include "instructions_bg.h"
#include "splash_bg.h"

#define SDK_DIM 9
#define SDK_PADDING 2
#define LINE_WIDTH 1
#define F_PADDING 4

#define SDK_WIDTH  (9 * (F_WIDTH  + F_PADDING + LINE_WIDTH) + 2 * SDK_PADDING)
#define SDK_HEIGHT (9 * (F_HEIGHT + F_PADDING + LINE_WIDTH) + 2 * SDK_PADDING)

#define C_BG         RGB(30,30,30)
#define C_SDK_BG     RGB(29,29,29)
#define C_SDK_BORDER RGB(25,25,25)
#define C_FONT       RGB(20,20,20)
#define C_FONT_SELECTED       RGB(10,20,10)
#define C_SDK_BG_SELECTED     RGB(28,31,28)
#define C_SDK_BORDER_SELECTED RGB(10,20,10)
#define C_SDK_BG_LOCKED     RGB(27,27,31)
#define C_SDK_BORDER_LOCKED RGB(15,15,15)
#define C_FONT_LOCKED       RGB(21,21,31)

typedef struct {
  int i;
  int j;
  int selected;
  int locked;
  int value;
  int answer;
} sdk_box_t;

typedef enum {SPLASH, INSTRUCTIONS, PLAY, WON, LOST} state_t;
typedef enum {FAILED, SUCCESS} result_t;

result_t play_sdk(uint16_t *vb);
result_t check_sdk(sdk_box_t grid[SDK_DIM][SDK_DIM]);
void init_sdk(sdk_box_t grid[SDK_DIM][SDK_DIM]);
void draw_sdk(uint16_t *vb, sdk_box_t grid[SDK_DIM][SDK_DIM]);
void draw_sdk_box(uint16_t *vb, sdk_box_t *box);

void play_splash(uint16_t *vb);
void play_instructions(uint16_t *vb);
void play_won(uint16_t *vb);
void play_lost(uint16_t *vb);

int main(void) {
  REG_DISPCTL = MODE3 | BG2_ENABLE;
  uint16_t *vb = VIDBUFADR;

  state_t game_state = SPLASH;

  while (1) {
    switch (game_state) {
      case SPLASH:
        //play_splash(vb); // Removing my face from the game... easter egg
        game_state = INSTRUCTIONS;
      break;
      case INSTRUCTIONS:
        play_instructions(vb);
        game_state = PLAY;
      break;
      case PLAY:
        game_state = (SUCCESS == play_sdk(vb)) ? WON : LOST;
      break;
      case WON:
        play_won(vb);
        game_state = SPLASH;
      break;
      case LOST:
        play_lost(vb);
        game_state = SPLASH;
      break;
    }
  }

  return 0;
}

result_t play_sdk(uint16_t *vb) {
  uint16_t last_buttons = BUTTONS;

  /* Sudoku grid data structure and its initialization */
  sdk_box_t grid[SDK_DIM][SDK_DIM];
  init_sdk(grid);

  sdk_box_t *selected = &grid[0][0];
  selected->selected = 1;

  /**
   * Create a pointer to the top left pixel of the Sudoku grid. This becomes 
   * convenient for referencing locations within the grid without having to 
   * apply the grid positioning algorithm again. The only disadvantage to this
   * method is it can overrun the buffer.
   */
  uint16_t *vb_sdk = &vb[PIXEL3((S_WIDTH  - SDK_WIDTH)/2,
                                (S_HEIGHT - SDK_HEIGHT)/2)];

  draw_image3(vb, play_bgBitmap);
  draw_sdk(vb_sdk, grid);
  draw_sdk_box(vb_sdk, selected);

  while (1) {
    while (SCANLINE_COUNTER < SCANLINE_OUT_OF_RANGE);

    /* Implement Pause... Stupid requirement */
    if (KEY_HIT(BUTTON_START, last_buttons) && !selected->locked) {
      draw_string3(vb_sdk, 19, 130, "Press Start", RGB(20,20,20));
      selected->locked = 1;
      draw_sdk_box(vb_sdk, selected);

      while(KEY_DOWN(BUTTON_START));
      while(!KEY_DOWN(BUTTON_START));

      draw_rect3(vb_sdk, 0, 130, 100, 140, WHITE);
      selected->locked = 0;
      draw_sdk_box(vb_sdk, selected);
    }

    /* If R or L is pressed increment/decrement */
    if (KEY_HIT(BUTTON_R, last_buttons) && !selected->locked) {
      selected->value = (selected->value + 1) % (SDK_DIM + 1);
      draw_sdk_box(vb_sdk, selected);
    }
    if (KEY_HIT(BUTTON_L, last_buttons) && !selected->locked) {
      /* To account for -1, add (SDK_DIM + 1) to the selected value */
      selected->value = (SDK_DIM + selected->value) % (SDK_DIM + 1);
      draw_sdk_box(vb_sdk, selected);
    }

    /* If the B key is pressed, clear the value */
    if (KEY_HIT(BUTTON_B, last_buttons) && !selected->locked) {
      selected->value = 0;
      draw_sdk_box(vb_sdk, selected);
    }
    /* If the A key is pressed, lock in the value */
    if (KEY_HIT(BUTTON_A, last_buttons) && selected->value 
                                        && !selected->locked) {
      selected->selected = 0;
      selected->locked   = 1;
      draw_sdk_box(vb_sdk, selected);

      if (FAILED == check_sdk(grid)) { return FAILED; }

      int i, j;
      for (j = 0; j < SDK_DIM; j++) {
        for (i = 0; i < SDK_DIM; i++) {
          if (!grid[i][j].locked && selected->locked) {
            selected = &grid[i][j];
            selected->selected = 1;
            draw_sdk_box(vb_sdk, selected);
          }
        }
      }
      if (selected->locked) { return SUCCESS; }
    }

    /* If a direction key is pressed */
    if (KEY_HIT(BUTTON_DOWN|BUTTON_UP|BUTTON_LEFT|BUTTON_RIGHT, last_buttons)) {
      int dir_i = 0, dir_j = 0;
      if (KEY_HIT(BUTTON_LEFT,  last_buttons)) { dir_i = -1; }
      if (KEY_HIT(BUTTON_RIGHT, last_buttons)) { dir_i =  1; }
      if (KEY_HIT(BUTTON_UP,    last_buttons)) { dir_j = -1; }
      if (KEY_HIT(BUTTON_DOWN,  last_buttons)) { dir_j =  1; }

      /* Unset previously selected box */
      selected->selected = 0;
      draw_sdk_box(vb_sdk, selected);

      /* While the direction applied to the canidate is within bounds */
      if ((selected->j + dir_j) >= 0 && (selected->j + dir_j) < SDK_DIM
       && (selected->i + dir_i) >= 0 && (selected->i + dir_i) < SDK_DIM) {

        selected = &grid[selected->i + dir_i][selected->j + dir_j];
      } 

      /* Draw new selection; this may be the previous selection */
      selected->selected = 1;
      draw_sdk_box(vb_sdk, selected);
    }

    last_buttons = BUTTONS;
  }

}

result_t check_sdk(sdk_box_t grid[SDK_DIM][SDK_DIM]) {
  int i, j;
  for (i=0; i<SDK_DIM; i++) {
    for (j=0; j<SDK_DIM; j++) {
      if (grid[i][j].value != grid[i][j].answer && grid[i][j].locked) {
        return FAILED;
      }
    }
  }
  return SUCCESS;
}

void init_sdk(sdk_box_t grid[SDK_DIM][SDK_DIM]) {
  int answers[9][9] = {
    {8,4,1,3,9,7,5,2,6},
    {3,6,2,8,4,5,1,7,9},
    {7,5,9,2,1,6,8,4,3},
    {2,3,7,5,6,8,4,9,1},
    {9,8,5,1,3,4,7,6,2},
    {4,1,6,9,7,2,3,5,8},
    {5,2,4,6,8,3,9,1,7},
    {6,9,3,7,5,1,2,8,4},
    {1,7,8,4,2,9,6,3,5}
  };
  int locked[9][9] = {
    {0,1,0,0,0,0,0,0,0},
    {1,1,1,0,0,0,0,0,0},
    {0,0,1,0,1,0,1,1,0},
    {0,1,0,0,0,1,0,1,0},
    {0,0,0,0,0,0,0,1,1},
    {0,1,1,1,1,0,0,0,0},
    {1,1,0,1,1,0,0,0,1},
    {1,0,1,0,0,0,1,0,0},
    {1,0,0,1,0,0,1,0,1}
  };

  int i, j;
  for (i=0; i<SDK_DIM; i++) {
    for (j=0; j<SDK_DIM; j++) {
      grid[i][j].selected = 0;
      grid[i][j].answer = answers[i][j];

      if (locked[i][j]) {
        grid[i][j].value  = grid[i][j].answer;
        grid[i][j].locked = 1;
      } else {
        grid[i][j].value  = 0;
        grid[i][j].locked = 0;
      }

      grid[i][j].i = i;
      grid[i][j].j = j;
    }
  }
}

void draw_sdk(uint16_t *vb_sdk, sdk_box_t grid[SDK_DIM][SDK_DIM]) {
  int i, j;

  draw_rect3(vb_sdk, 0, 0, SDK_WIDTH, SDK_HEIGHT, C_SDK_BG);

  /* Draw the box's left and top lines skipping a bit every third box. */
  for (i = 0; i < SDK_DIM; i++) {
    int x = i * (F_WIDTH  + F_PADDING + LINE_WIDTH) + SDK_PADDING * (i/3);
    int y = i * (F_HEIGHT + F_PADDING + LINE_WIDTH) + SDK_PADDING * (i/3);
    draw_line3(vb_sdk, x, 0, x,         SDK_HEIGHT, C_SDK_BORDER);
    draw_line3(vb_sdk, 0, y, SDK_WIDTH, y,          C_SDK_BORDER);
  }

  /* Draw the box's remaining right and bottom lines. */
  for (i = SDK_DIM/3; i <= SDK_DIM; i += SDK_DIM/3) {
    int x = i * (F_WIDTH  + F_PADDING + LINE_WIDTH) + SDK_PADDING * ((i-1)/3);
    int y = i * (F_HEIGHT + F_PADDING + LINE_WIDTH) + SDK_PADDING * ((i-1)/3);
    draw_line3(vb_sdk, x, 0, x,         SDK_HEIGHT, C_SDK_BORDER);
    draw_line3(vb_sdk, 0, y, SDK_WIDTH, y,          C_SDK_BORDER);
  }

  /* Break apart the 3x3 boxes */
  for (i = SDK_DIM/3; i <= SDK_DIM; i += SDK_DIM/3) {
    int x = i * (F_WIDTH  + F_PADDING + LINE_WIDTH) + SDK_PADDING * ((i-1)/3);
    int y = i * (F_HEIGHT + F_PADDING + LINE_WIDTH) + SDK_PADDING * ((i-1)/3);
    x++; y++;
    draw_line3(vb_sdk, x, 0, x,         SDK_HEIGHT, C_BG);
    draw_line3(vb_sdk, 0, y, SDK_WIDTH, y,          C_BG);
  }

  /* Draw the boxs that have been initialized with some value */
  for (i=0; i<SDK_DIM; i++) {
    for (j=0; j<SDK_DIM; j++) {
      if (grid[i][j].value || grid[i][j].selected) {
        draw_sdk_box(vb_sdk, &grid[i][j]);
      }
    }
  }
}

void draw_sdk_box(uint16_t *vb_sdk, sdk_box_t *box) {
  /**
   * Need to find the top left of the box for convient reference. As mentioned
   * with the vb_sdk offset, the standard functions will not prevent buffer over
   * runs when an offset is used.
   */
  int x = box->i*(F_WIDTH  + F_PADDING + LINE_WIDTH) + SDK_PADDING*(box->i/3);
  int y = box->j*(F_HEIGHT + F_PADDING + LINE_WIDTH) + SDK_PADDING*(box->j/3);
  uint16_t *vb_box = &vb_sdk[PIXEL3(x,y)];

  uint16_t border = C_SDK_BORDER;
  uint16_t bg     = C_SDK_BG;
  uint16_t fg     = C_FONT;

  /* Selected and locked use a different set of colors */
  if (box->locked && box->selected) {
    border = C_SDK_BORDER_LOCKED;
    bg = C_SDK_BG_LOCKED;
    fg = C_FONT_LOCKED;

  } else if (box->locked) {
    bg = C_SDK_BG_LOCKED;
    fg = C_FONT_LOCKED;

  } else if (box->selected) {
    border = C_SDK_BORDER_SELECTED;
    bg     = C_SDK_BG_SELECTED;
    fg     = C_FONT_SELECTED;
  }

  /* Locate the boundries based on drawing rules specified in draw_sdk */
  int outer_x0 = 0;
  int outer_y0 = 0;
  int outer_x1 = F_WIDTH  + F_PADDING + 2*LINE_WIDTH;
  int outer_y1 = F_HEIGHT + F_PADDING + 2*LINE_WIDTH - 1;

  int inner_x0 = outer_x0 + LINE_WIDTH;
  int inner_y0 = outer_y0 + LINE_WIDTH;
  int inner_x1 = outer_x1 - LINE_WIDTH;
  int inner_y1 = outer_y1 - LINE_WIDTH;

  int offset_x = F_PADDING/2 + LINE_WIDTH - 1;
  int offset_y = F_PADDING/2 + LINE_WIDTH;

  draw_rect3(vb_box, outer_x0, outer_y0, outer_x1, outer_y1, border);
  draw_rect3(vb_box, inner_x0, inner_y0, inner_x1, inner_y1, bg);

  if (box->value) draw_char3(vb_box, offset_x, offset_y, box->value+'0', fg);
}

void play_splash(uint16_t *vb) {
  draw_image3(vb, splash_bgBitmap);
  while(!KEY_DOWN(BUTTON_START));
}
void play_instructions(uint16_t *vb) {
  draw_image3(vb, instructions_bgBitmap);
  while(KEY_DOWN(BUTTON_START));
  while(!KEY_DOWN(BUTTON_START));
}
void play_won(uint16_t *vb) {
  draw_string3(vb, 66, 150, "Great! Press start", C_SDK_BORDER_SELECTED);
  while(!KEY_DOWN(BUTTON_START));
}
void play_lost(uint16_t *vb) {
  draw_string3(vb, 69, 150, "FAIL! Press Start", RGB(31,10,10));
  while(!KEY_DOWN(BUTTON_START));
}
