/**
 * Shape drawing functions for Mode 3, Background 2
 *
 * @author Ben Heiskell
 * @date 2009 08
 */

#include "gba.h"
#include "shape.h"

/**
 * Draw a fullscreen image.
 */
inline void draw_image3(uint16_t *video_buffer, const unsigned short *image) {
    DMA[3].src = (void *)image;
    DMA[3].dst = (void *)video_buffer;
    DMA[3].cnt = (S_WIDTH * S_HEIGHT) | DMA_ON;
}

/**
 * Draw a pixel with buffer overrunning checking.
 */
inline void draw_pixel3(uint16_t *video_buffer, int x, int y, uint16_t color) {
  if (x >= 0 && x < S_WIDTH && y >= 0 && y < S_WIDTH) {
    video_buffer[PIXEL3(x,y)] = color;
  }
}

/**
 * Draw a line from point X to point Y
 * Based on Bresenham's Algorithm - Wikipedia FTW
 */
void draw_line3(uint16_t *video_buffer, int x1, int y1, int x2, int y2, 
                uint16_t color) {

  int diff_y = y2 - y1;
  int diff_x = x2 - x1;
  if (0 > diff_y) { diff_y = -diff_y; }
  if (0 > diff_x) { diff_x = -diff_x; }

  int swap, steep = (diff_y > diff_x) ? 1 : 0;

  if (steep)   {
    swap = x1; x1 = y1; y1 = swap;
    swap = x2; x2 = y2; y2 = swap;
    swap = diff_x; diff_x = diff_y; diff_y = swap;
  }
  if (x1 > x2) {
    swap = x1; x1 = x2; x2 = swap;
    swap = y1; y1 = y2; y2 = swap;
  }

  double error = 0;
  int x, y = y1, y_inc = (y1 < y2) ? 1 : -1;

  for (x = x1; x <= x2; x++) {
    if (steep) { 
      draw_pixel3(video_buffer, y, x, color);
    } else {
      draw_pixel3(video_buffer, x, y, color);
    }

    error += diff_y / (double)diff_x;
    if (error >= 0.5) { y += y_inc; error--; }
  }
}

/**
 * Draw a rectangle from one point to another.
 */
void draw_rect3(uint16_t *video_buffer, int x1, int y1, int x2, int y2,
                uint16_t color) {
  int swap, y;

  /* Ensure we're counting upwards */
  if (x1 > x2) { swap = x1; x1 = x2; x2 = swap; }
  if (y1 > y2) { swap = y1; y1 = y2; y2 = swap; }

  for (y = y1; y <= y2; y++) {
    DMA[3].src = &color;
    DMA[3].dst = &video_buffer[PIXEL3(x1, y)];
    DMA[3].cnt = (x2 - x1) | DMA_SOURCE_FIXED | DMA_ON;
  }
}

/**
 * Draw the error box and lock up the program.
 */
void draw_error3(uint16_t *video_buffer) {
  draw_line3(video_buffer, 0,         0,          0,         S_HEIGHT-1, RED);
  draw_line3(video_buffer, 0,         0,          S_WIDTH-1, 0,          RED);
  draw_line3(video_buffer, 0,         S_HEIGHT-1, S_WIDTH-1, S_HEIGHT-1, RED);
  draw_line3(video_buffer, S_WIDTH-1, 0,          S_WIDTH-1, S_HEIGHT-1, RED);
  while (1);
}
